@@ -1,4 +1,4 @@
-# Copyright 2014-2015 MongoDB, Inc.
+# Copyright 2014-present MongoDB, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -16,22 +16,30 @@
 
 .. versionadded:: 2.7
 """
+import copy
+
+from itertools import islice
 
 from bson.objectid import ObjectId
 from bson.raw_bson import RawBSONDocument
 from bson.son import SON
+from pymongo.client_session import _validate_session_write_concern
 from pymongo.common import (validate_is_mapping,
                             validate_is_document_type,
                             validate_ok_for_replace,
                             validate_ok_for_update)
+from pymongo.helpers import _RETRYABLE_ERROR_CODES
+from pymongo.collation import validate_collation_or_none
 from pymongo.errors import (BulkWriteError,
-                            DocumentTooLarge,
+                            ConfigurationError,
                             InvalidOperation,
                             OperationFailure)
 from pymongo.message import (_INSERT, _UPDATE, _DELETE,
-                             _do_batched_write_command,
+                             _do_batched_insert,
                              _randint,
-                             _BulkWriteContext)
+                             _BulkWriteContext,
+                             _EncryptedBulkWriteContext)
+from pymongo.read_preferences import ReadPreference
 from pymongo.write_concern import WriteConcern
 
 
@@ -49,10 +57,6 @@ _COMMANDS = ('insert', 'update', 'delete')
 # These string literals are used when we create fake server return
 # documents client side. We use unicode literals in python 2.x to
 # match the actual return values from the server.
-_UID = u"_id"
-_UCODE = u"code"
-_UERRMSG = u"errmsg"
-_UINDEX = u"index"
 _UOP = u"op"
 
 
@@ -65,6 +69,7 @@ class _Run(object):
         self.op_type = op_type
         self.index_map = []
         self.ops = []
+        self.idx_offset = 0
 
     def index(self, idx):
         """Get the original index of an operation in this run.
@@ -86,114 +91,53 @@ class _Run(object):
         self.ops.append(operation)
 
 
-def _make_error(index, code, errmsg, operation):
-    """Create and return an error document.
+def _merge_command(run, full_result, offset, result):
+    """Merge a write command result into the full bulk result.
     """
-    return {
-        _UINDEX: index,
-        _UCODE: code,
-        _UERRMSG: errmsg,
-        _UOP: operation
-    }
+    affected = result.get("n", 0)
 
-
-def _merge_legacy(run, full_result, result, index):
-    """Merge a result from a legacy opcode into the full results.
-    """
-    affected = result.get('n', 0)
-
-    errmsg = result.get("errmsg", result.get("err", ""))
-    if errmsg:
-        # wtimeout is not considered a hard failure in
-        # MongoDB 2.6 so don't treat it like one here.
-        if result.get("wtimeout"):
-            error_doc = {'errmsg': errmsg, 'code': _WRITE_CONCERN_ERROR}
-            full_result['writeConcernErrors'].append(error_doc)
-        else:
-            code = result.get("code", _UNKNOWN_ERROR)
-            error = _make_error(run.index(index), code, errmsg, run.ops[index])
-            if "errInfo" in result:
-                error["errInfo"] = result["errInfo"]
-            full_result["writeErrors"].append(error)
-            return
     if run.op_type == _INSERT:
-        full_result['nInserted'] += 1
-    elif run.op_type == _UPDATE:
-        if "upserted" in result:
-            doc = {_UINDEX: run.index(index), _UID: result["upserted"]}
-            full_result["upserted"].append(doc)
-            full_result['nUpserted'] += affected
-        # Versions of MongoDB before 2.6 don't return the _id for an
-        # upsert if _id is not an ObjectId.
-        elif result.get("updatedExisting") is False and affected == 1:
-            op = run.ops[index]
-            # If _id is in both the update document *and* the query spec
-            # the update document _id takes precedence.
-            _id = op['u'].get('_id', op['q'].get('_id'))
-            doc = {_UINDEX: run.index(index), _UID: _id}
-            full_result["upserted"].append(doc)
-            full_result['nUpserted'] += affected
-        else:
-            full_result['nMatched'] += affected
+        full_result["nInserted"] += affected
 
     elif run.op_type == _DELETE:
-        full_result['nRemoved'] += affected
+        full_result["nRemoved"] += affected
+
+    elif run.op_type == _UPDATE:
+        upserted = result.get("upserted")
+        if upserted:
+            n_upserted = len(upserted)
+            for doc in upserted:
+                doc["index"] = run.index(doc["index"] + offset)
+            full_result["upserted"].extend(upserted)
+            full_result["nUpserted"] += n_upserted
+            full_result["nMatched"] += (affected - n_upserted)
+        else:
+            full_result["nMatched"] += affected
+        full_result["nModified"] += result["nModified"]
+
+    write_errors = result.get("writeErrors")
+    if write_errors:
+        for doc in write_errors:
+            # Leave the server response intact for APM.
+            replacement = doc.copy()
+            idx = doc["index"] + offset
+            replacement["index"] = run.index(idx)
+            # Add the failed operation to the error document.
+            replacement[_UOP] = run.ops[idx]
+            full_result["writeErrors"].append(replacement)
+
+    wc_error = result.get("writeConcernError")
+    if wc_error:
+        full_result["writeConcernErrors"].append(wc_error)
 
 
-def _merge_command(run, full_result, results):
-    """Merge a group of results from write commands into the full result.
+def _raise_bulk_write_error(full_result):
+    """Raise a BulkWriteError from the full bulk api result.
     """
-    for offset, result in results:
-
-        affected = result.get("n", 0)
-
-        if run.op_type == _INSERT:
-            full_result["nInserted"] += affected
-
-        elif run.op_type == _DELETE:
-            full_result["nRemoved"] += affected
-
-        elif run.op_type == _UPDATE:
-            upserted = result.get("upserted")
-            if upserted:
-                if isinstance(upserted, list):
-                    n_upserted = len(upserted)
-                    for doc in upserted:
-                        doc["index"] = run.index(doc["index"] + offset)
-                    full_result["upserted"].extend(upserted)
-                else:
-                    n_upserted = 1
-                    index = run.index(offset)
-                    doc = {_UINDEX: index, _UID: upserted}
-                    full_result["upserted"].append(doc)
-                full_result["nUpserted"] += n_upserted
-                full_result["nMatched"] += (affected - n_upserted)
-            else:
-                full_result["nMatched"] += affected
-            n_modified = result.get("nModified")
-            # SERVER-13001 - in a mixed sharded cluster a call to
-            # update could return nModified (>= 2.6) or not (<= 2.4).
-            # If any call does not return nModified we can't report
-            # a valid final count so omit the field completely.
-            if n_modified is not None and "nModified" in full_result:
-                full_result["nModified"] += n_modified
-            else:
-                full_result.pop("nModified", None)
-
-        write_errors = result.get("writeErrors")
-        if write_errors:
-            for doc in write_errors:
-                # Leave the server response intact for APM.
-                replacement = doc.copy()
-                idx = doc["index"] + offset
-                replacement["index"] = run.index(idx)
-                # Add the failed operation to the error document.
-                replacement[_UOP] = run.ops[idx]
-                full_result["writeErrors"].append(replacement)
-
-        wc_error = result.get("writeConcernError")
-        if wc_error:
-            full_result["writeConcernErrors"].append(wc_error)
+    if full_result["writeErrors"]:
+        full_result["writeErrors"].sort(
+            key=lambda error: error["index"])
+    raise BulkWriteError(full_result)
 
 
 class _Bulk(object):
@@ -208,10 +152,24 @@ class _Bulk(object):
                 document_class=dict))
         self.ordered = ordered
         self.ops = []
-        self.name = "%s.%s" % (collection.database.name, collection.name)
-        self.namespace = collection.database.name + '.$cmd'
         self.executed = False
         self.bypass_doc_val = bypass_document_validation
+        self.uses_collation = False
+        self.uses_array_filters = False
+        self.uses_hint = False
+        self.is_retryable = True
+        self.retrying = False
+        self.started_retryable_write = False
+        # Extra state so that we know where to pick up on a retry attempt.
+        self.current_run = None
+
+    @property
+    def bulk_ctx_class(self):
+        encrypter = self.collection.database.client._encrypter
+        if encrypter and not encrypter._bypass_auto_encryption:
+            return _EncryptedBulkWriteContext
+        else:
+            return _BulkWriteContext
 
     def add_insert(self, document):
         """Add an insert document to the list of ops.
@@ -222,26 +180,58 @@ class _Bulk(object):
             document['_id'] = ObjectId()
         self.ops.append((_INSERT, document))
 
-    def add_update(self, selector, update, multi=False, upsert=False):
+    def add_update(self, selector, update, multi=False, upsert=False,
+                   collation=None, array_filters=None, hint=None):
         """Create an update document and add it to the list of ops.
         """
         validate_ok_for_update(update)
         cmd = SON([('q', selector), ('u', update),
                    ('multi', multi), ('upsert', upsert)])
+        collation = validate_collation_or_none(collation)
+        if collation is not None:
+            self.uses_collation = True
+            cmd['collation'] = collation
+        if array_filters is not None:
+            self.uses_array_filters = True
+            cmd['arrayFilters'] = array_filters
+        if hint is not None:
+            self.uses_hint = True
+            cmd['hint'] = hint
+        if multi:
+            # A bulk_write containing an update_many is not retryable.
+            self.is_retryable = False
         self.ops.append((_UPDATE, cmd))
 
-    def add_replace(self, selector, replacement, upsert=False):
+    def add_replace(self, selector, replacement, upsert=False,
+                    collation=None, hint=None):
         """Create a replace document and add it to the list of ops.
         """
         validate_ok_for_replace(replacement)
         cmd = SON([('q', selector), ('u', replacement),
                    ('multi', False), ('upsert', upsert)])
+        collation = validate_collation_or_none(collation)
+        if collation is not None:
+            self.uses_collation = True
+            cmd['collation'] = collation
+        if hint is not None:
+            self.uses_hint = True
+            cmd['hint'] = hint
         self.ops.append((_UPDATE, cmd))
 
-    def add_delete(self, selector, limit):
+    def add_delete(self, selector, limit, collation=None, hint=None):
         """Create a delete document and add it to the list of ops.
         """
         cmd = SON([('q', selector), ('limit', limit)])
+        collation = validate_collation_or_none(collation)
+        if collation is not None:
+            self.uses_collation = True
+            cmd['collation'] = collation
+        if hint is not None:
+            self.uses_hint = True
+            cmd['hint'] = hint
+        if limit == _DELETE_ALL:
+            # A bulk_write containing a delete_many is not retryable.
+            self.is_retryable = False
         self.ops.append((_DELETE, cmd))
 
     def gen_ordered(self):
@@ -270,7 +260,81 @@ class _Bulk(object):
             if run.ops:
                 yield run
 
-    def execute_command(self, sock_info, generator, write_concern):
+    def _execute_command(self, generator, write_concern, session,
+                         sock_info, op_id, retryable, full_result):
+        if sock_info.max_wire_version < 5:
+            if self.uses_collation:
+                raise ConfigurationError(
+                    'Must be connected to MongoDB 3.4+ to use a collation.')
+            if self.uses_hint:
+                raise ConfigurationError(
+                    'Must be connected to MongoDB 3.4+ to use hint.')
+        if sock_info.max_wire_version < 6 and self.uses_array_filters:
+            raise ConfigurationError(
+                'Must be connected to MongoDB 3.6+ to use arrayFilters.')
+
+        db_name = self.collection.database.name
+        client = self.collection.database.client
+        listeners = client._event_listeners
+
+        if not self.current_run:
+            self.current_run = next(generator)
+        run = self.current_run
+
+        # sock_info.command validates the session, but we use
+        # sock_info.write_command.
+        sock_info.validate_session(client, session)
+        while run:
+            cmd = SON([(_COMMANDS[run.op_type], self.collection.name),
+                       ('ordered', self.ordered)])
+            if not write_concern.is_server_default:
+                cmd['writeConcern'] = write_concern.document
+            if self.bypass_doc_val and sock_info.max_wire_version >= 4:
+                cmd['bypassDocumentValidation'] = True
+            bwc = self.bulk_ctx_class(
+                db_name, cmd, sock_info, op_id, listeners, session,
+                run.op_type, self.collection.codec_options)
+
+            while run.idx_offset < len(run.ops):
+                if session:
+                    # Start a new retryable write unless one was already
+                    # started for this command.
+                    if retryable and not self.started_retryable_write:
+                        session._start_retryable_write()
+                        self.started_retryable_write = True
+                    session._apply_to(cmd, retryable, ReadPreference.PRIMARY,
+                                      sock_info)
+                sock_info.send_cluster_time(cmd, session, client)
+                ops = islice(run.ops, run.idx_offset, None)
+                # Run as many ops as possible in one command.
+                result, to_send = bwc.execute(ops, client)
+
+                # Retryable writeConcernErrors halt the execution of this run.
+                wce = result.get('writeConcernError', {})
+                if wce.get('code', 0) in _RETRYABLE_ERROR_CODES:
+                    # Synthesize the full bulk result without modifying the
+                    # current one because this write operation may be retried.
+                    full = copy.deepcopy(full_result)
+                    _merge_command(run, full, run.idx_offset, result)
+                    _raise_bulk_write_error(full)
+
+                _merge_command(run, full_result, run.idx_offset, result)
+                # We're no longer in a retry once a command succeeds.
+                self.retrying = False
+                self.started_retryable_write = False
+
+                if self.ordered and "writeErrors" in result:
+                    break
+                run.idx_offset += len(to_send)
+
+            # We're supposed to continue if errors are
+            # at the write concern level (e.g. wtimeout)
+            if self.ordered and full_result['writeErrors']:
+                break
+            # Reset our state
+            self.current_run = run = next(generator, None)
+
+    def execute_command(self, generator, write_concern, session):
         """Execute using write commands.
         """
         # nModified is only reported for write commands, not legacy ops.
@@ -285,58 +349,132 @@ class _Bulk(object):
             "upserted": [],
         }
         op_id = _randint()
-        db_name = self.collection.database.name
-        listeners = self.collection.database.client._event_listeners
 
-        for run in generator:
-            cmd = SON([(_COMMANDS[run.op_type], self.collection.name),
-                       ('ordered', self.ordered)])
-            if write_concern.document:
-                cmd['writeConcern'] = write_concern.document
-            if self.bypass_doc_val and sock_info.max_wire_version >= 4:
-                cmd['bypassDocumentValidation'] = True
+        def retryable_bulk(session, sock_info, retryable):
+            self._execute_command(
+                generator, write_concern, session, sock_info, op_id,
+                retryable, full_result)
 
-            bwc = _BulkWriteContext(db_name, cmd, sock_info, op_id, listeners)
-            results = _do_batched_write_command(
-                self.namespace, run.op_type, cmd,
-                run.ops, True, self.collection.codec_options, bwc)
-
-            _merge_command(run, full_result, results)
-            # We're supposed to continue if errors are
-            # at the write concern level (e.g. wtimeout)
-            if self.ordered and full_result['writeErrors']:
-                break
+        client = self.collection.database.client
+        with client._tmp_session(session) as s:
+            client._retry_with_session(
+                self.is_retryable, retryable_bulk, s, self)
 
         if full_result["writeErrors"] or full_result["writeConcernErrors"]:
-            if full_result['writeErrors']:
-                full_result['writeErrors'].sort(
-                    key=lambda error: error['index'])
-            raise BulkWriteError(full_result)
+            _raise_bulk_write_error(full_result)
         return full_result
 
+    def execute_insert_no_results(self, sock_info, run, op_id, acknowledged):
+        """Execute insert, returning no results.
+        """
+        command = SON([('insert', self.collection.name),
+                       ('ordered', self.ordered)])
+        concern = {'w': int(self.ordered)}
+        command['writeConcern'] = concern
+        if self.bypass_doc_val and sock_info.max_wire_version >= 4:
+            command['bypassDocumentValidation'] = True
+        db = self.collection.database
+        bwc = _BulkWriteContext(
+            db.name, command, sock_info, op_id, db.client._event_listeners,
+            None, _INSERT, self.collection.codec_options)
+        # Legacy batched OP_INSERT.
+        _do_batched_insert(
+            self.collection.full_name, run.ops, True, acknowledged, concern,
+            not self.ordered, self.collection.codec_options, bwc)
+
+    def execute_op_msg_no_results(self, sock_info, generator):
+        """Execute write commands with OP_MSG and w=0 writeConcern, unordered.
+        """
+        db_name = self.collection.database.name
+        client = self.collection.database.client
+        listeners = client._event_listeners
+        op_id = _randint()
+
+        if not self.current_run:
+            self.current_run = next(generator)
+        run = self.current_run
+
+        while run:
+            cmd = SON([(_COMMANDS[run.op_type], self.collection.name),
+                       ('ordered', False),
+                       ('writeConcern', {'w': 0})])
+            bwc = self.bulk_ctx_class(
+                db_name, cmd, sock_info, op_id, listeners, None,
+                run.op_type, self.collection.codec_options)
+
+            while run.idx_offset < len(run.ops):
+                ops = islice(run.ops, run.idx_offset, None)
+                # Run as many ops as possible.
+                to_send = bwc.execute_unack(ops, client)
+                run.idx_offset += len(to_send)
+            self.current_run = run = next(generator, None)
+
+    def execute_command_no_results(self, sock_info, generator):
+        """Execute write commands with OP_MSG and w=0 WriteConcern, ordered.
+        """
+        full_result = {
+            "writeErrors": [],
+            "writeConcernErrors": [],
+            "nInserted": 0,
+            "nUpserted": 0,
+            "nMatched": 0,
+            "nModified": 0,
+            "nRemoved": 0,
+            "upserted": [],
+        }
+        # Ordered bulk writes have to be acknowledged so that we stop
+        # processing at the first error, even when the application
+        # specified unacknowledged writeConcern.
+        write_concern = WriteConcern()
+        op_id = _randint()
+        try:
+            self._execute_command(
+                generator, write_concern, None,
+                sock_info, op_id, False, full_result)
+        except OperationFailure:
+            pass
+
     def execute_no_results(self, sock_info, generator):
         """Execute all operations, returning no results (w=0).
         """
-        # Cannot have both unacknowledged write and bypass document validation.
+        if self.uses_collation:
+            raise ConfigurationError(
+                'Collation is unsupported for unacknowledged writes.')
+        if self.uses_array_filters:
+            raise ConfigurationError(
+                'arrayFilters is unsupported for unacknowledged writes.')
+        if self.uses_hint:
+            raise ConfigurationError(
+                'hint is unsupported for unacknowledged writes.')
+        # Cannot have both unacknowledged writes and bypass document validation.
         if self.bypass_doc_val and sock_info.max_wire_version >= 4:
             raise OperationFailure("Cannot set bypass_document_validation with"
                                    " unacknowledged write concern")
+
+        # OP_MSG
+        if sock_info.max_wire_version > 5:
+            if self.ordered:
+                return self.execute_command_no_results(sock_info, generator)
+            return self.execute_op_msg_no_results(sock_info, generator)
+
         coll = self.collection
         # If ordered is True we have to send GLE or use write
         # commands so we can abort on the first error.
         write_concern = WriteConcern(w=int(self.ordered))
         op_id = _randint()
 
-        for run in generator:
+        next_run = next(generator)
+        while next_run:
+            # An ordered bulk write needs to send acknowledged writes to short
+            # circuit the next run. However, the final message on the final
+            # run can be unacknowledged.
+            run = next_run
+            next_run = next(generator, None)
+            needs_ack = self.ordered and next_run is not None
             try:
                 if run.op_type == _INSERT:
-                    coll._insert(
-                        sock_info,
-                        run.ops,
-                        self.ordered,
-                        write_concern=write_concern,
-                        op_id=op_id,
-                        bypass_doc_val=self.bypass_doc_val)
+                    self.execute_insert_no_results(
+                        sock_info, run, op_id, needs_ack)
                 elif run.op_type == _UPDATE:
                     for operation in run.ops:
                         doc = operation['u']
@@ -366,86 +504,7 @@ class _Bulk(object):
                 if self.ordered:
                     break
 
-    def execute_legacy(self, sock_info, generator, write_concern):
-        """Execute using legacy wire protocol ops.
-        """
-        coll = self.collection
-        full_result = {
-            "writeErrors": [],
-            "writeConcernErrors": [],
-            "nInserted": 0,
-            "nUpserted": 0,
-            "nMatched": 0,
-            "nRemoved": 0,
-            "upserted": [],
-        }
-        op_id = _randint()
-        stop = False
-        for run in generator:
-            for idx, operation in enumerate(run.ops):
-                try:
-                    # To do per-operation reporting we have to do ops one
-                    # at a time. That means the performance of bulk insert
-                    # will be slower here than calling Collection.insert()
-                    if run.op_type == _INSERT:
-                        coll._insert(sock_info,
-                                     operation,
-                                     self.ordered,
-                                     write_concern=write_concern,
-                                     op_id=op_id)
-                        result = {}
-                    elif run.op_type == _UPDATE:
-                        doc = operation['u']
-                        check_keys = True
-                        if doc and next(iter(doc)).startswith('$'):
-                            check_keys = False
-                        result = coll._update(sock_info,
-                                              operation['q'],
-                                              doc,
-                                              operation['upsert'],
-                                              check_keys,
-                                              operation['multi'],
-                                              write_concern=write_concern,
-                                              op_id=op_id,
-                                              ordered=self.ordered)
-                    else:
-                        result = coll._delete(sock_info,
-                                              operation['q'],
-                                              not operation['limit'],
-                                              write_concern,
-                                              op_id,
-                                              self.ordered)
-                    _merge_legacy(run, full_result, result, idx)
-                except DocumentTooLarge as exc:
-                    # MongoDB 2.6 uses error code 2 for "too large".
-                    error = _make_error(
-                        run.index(idx), _BAD_VALUE, str(exc), operation)
-                    full_result['writeErrors'].append(error)
-                    if self.ordered:
-                        stop = True
-                        break
-                except OperationFailure as exc:
-                    if not exc.details:
-                        # Some error not related to the write operation
-                        # (e.g. kerberos failure). Re-raise immediately.
-                        raise
-                    _merge_legacy(run, full_result, exc.details, idx)
-                    # We're supposed to continue if errors are
-                    # at the write concern level (e.g. wtimeout)
-                    if self.ordered and full_result["writeErrors"]:
-                        stop = True
-                        break
-            if stop:
-                break
-
-        if full_result["writeErrors"] or full_result['writeConcernErrors']:
-            if full_result['writeErrors']:
-                full_result['writeErrors'].sort(
-                    key=lambda error: error['index'])
-            raise BulkWriteError(full_result)
-        return full_result
-
-    def execute(self, write_concern):
+    def execute(self, write_concern, session):
         """Execute operations.
         """
         if not self.ops:
@@ -454,8 +513,8 @@ class _Bulk(object):
             raise InvalidOperation('Bulk operations can '
                                    'only be executed once.')
         self.executed = True
-        write_concern = (WriteConcern(**write_concern) if
-                         write_concern else self.collection.write_concern)
+        write_concern = write_concern or self.collection.write_concern
+        session = _validate_session_write_concern(session, write_concern)
 
         if self.ordered:
             generator = self.gen_ordered()
@@ -463,24 +522,23 @@ class _Bulk(object):
             generator = self.gen_unordered()
 
         client = self.collection.database.client
-        with client._socket_for_writes() as sock_info:
-            if not write_concern.acknowledged:
+        if not write_concern.acknowledged:
+            with client._socket_for_writes(session) as sock_info:
                 self.execute_no_results(sock_info, generator)
-            elif sock_info.max_wire_version > 1:
-                return self.execute_command(sock_info, generator, write_concern)
-            else:
-                return self.execute_legacy(sock_info, generator, write_concern)
+        else:
+            return self.execute_command(generator, write_concern, session)
 
 
 class BulkUpsertOperation(object):
     """An interface for adding upsert operations.
     """
 
-    __slots__ = ('__selector', '__bulk')
+    __slots__ = ('__selector', '__bulk', '__collation')
 
-    def __init__(self, selector, bulk):
+    def __init__(self, selector, bulk, collation):
         self.__selector = selector
         self.__bulk = bulk
+        self.__collation = collation
 
     def update_one(self, update):
         """Update one document matching the selector.
@@ -489,7 +547,8 @@ class BulkUpsertOperation(object):
           - `update` (dict): the update operations to apply
         """
         self.__bulk.add_update(self.__selector,
-                               update, multi=False, upsert=True)
+                               update, multi=False, upsert=True,
+                               collation=self.__collation)
 
     def update(self, update):
         """Update all documents matching the selector.
@@ -498,7 +557,8 @@ class BulkUpsertOperation(object):
           - `update` (dict): the update operations to apply
         """
         self.__bulk.add_update(self.__selector,
-                               update, multi=True, upsert=True)
+                               update, multi=True, upsert=True,
+                               collation=self.__collation)
 
     def replace_one(self, replacement):
         """Replace one entire document matching the selector criteria.
@@ -506,18 +566,20 @@ class BulkUpsertOperation(object):
         :Parameters:
           - `replacement` (dict): the replacement document
         """
-        self.__bulk.add_replace(self.__selector, replacement, upsert=True)
+        self.__bulk.add_replace(self.__selector, replacement, upsert=True,
+                                collation=self.__collation)
 
 
 class BulkWriteOperation(object):
     """An interface for adding update or remove operations.
     """
 
-    __slots__ = ('__selector', '__bulk')
+    __slots__ = ('__selector', '__bulk', '__collation')
 
-    def __init__(self, selector, bulk):
+    def __init__(self, selector, bulk, collation):
         self.__selector = selector
         self.__bulk = bulk
+        self.__collation = collation
 
     def update_one(self, update):
         """Update one document matching the selector criteria.
@@ -525,7 +587,8 @@ class BulkWriteOperation(object):
         :Parameters:
           - `update` (dict): the update operations to apply
         """
-        self.__bulk.add_update(self.__selector, update, multi=False)
+        self.__bulk.add_update(self.__selector, update, multi=False,
+                               collation=self.__collation)
 
     def update(self, update):
         """Update all documents matching the selector criteria.
@@ -533,7 +596,8 @@ class BulkWriteOperation(object):
         :Parameters:
           - `update` (dict): the update operations to apply
         """
-        self.__bulk.add_update(self.__selector, update, multi=True)
+        self.__bulk.add_update(self.__selector, update, multi=True,
+                               collation=self.__collation)
 
     def replace_one(self, replacement):
         """Replace one entire document matching the selector criteria.
@@ -541,17 +605,20 @@ class BulkWriteOperation(object):
         :Parameters:
           - `replacement` (dict): the replacement document
         """
-        self.__bulk.add_replace(self.__selector, replacement)
+        self.__bulk.add_replace(self.__selector, replacement,
+                                collation=self.__collation)
 
     def remove_one(self):
         """Remove a single document matching the selector criteria.
         """
-        self.__bulk.add_delete(self.__selector, _DELETE_ONE)
+        self.__bulk.add_delete(self.__selector, _DELETE_ONE,
+                               collation=self.__collation)
 
     def remove(self):
         """Remove all documents matching the selector criteria.
         """
-        self.__bulk.add_delete(self.__selector, _DELETE_ALL)
+        self.__bulk.add_delete(self.__selector, _DELETE_ALL,
+                               collation=self.__collation)
 
     def upsert(self):
         """Specify that all chained update operations should be
@@ -561,18 +628,19 @@ class BulkWriteOperation(object):
           - A :class:`BulkUpsertOperation` instance, used to add
             update operations to this bulk operation.
         """
-        return BulkUpsertOperation(self.__selector, self.__bulk)
+        return BulkUpsertOperation(self.__selector, self.__bulk,
+                                   self.__collation)
 
 
 class BulkOperationBuilder(object):
-    """An interface for executing a batch of write operations.
+    """**DEPRECATED**: An interface for executing a batch of write operations.
     """
 
     __slots__ = '__bulk'
 
     def __init__(self, collection, ordered=True,
                  bypass_document_validation=False):
-        """Initialize a new BulkOperationBuilder instance.
+        """**DEPRECATED**: Initialize a new BulkOperationBuilder instance.
 
         :Parameters:
           - `collection`: A :class:`~pymongo.collection.Collection` instance.
@@ -589,24 +657,35 @@ class BulkOperationBuilder(object):
         .. note:: `bypass_document_validation` requires server version
           **>= 3.2**
 
+        .. versionchanged:: 3.5
+           Deprecated. Use :meth:`~pymongo.collection.Collection.bulk_write`
+           instead.
+
         .. versionchanged:: 3.2
           Added bypass_document_validation support
         """
         self.__bulk = _Bulk(collection, ordered, bypass_document_validation)
 
-    def find(self, selector):
+    def find(self, selector, collation=None):
         """Specify selection criteria for bulk operations.
 
         :Parameters:
           - `selector` (dict): the selection criteria for update
             and remove operations.
+          - `collation` (optional): An instance of
+            :class:`~pymongo.collation.Collation`. This option is only
+            supported on MongoDB 3.4 and above.
 
         :Returns:
           - A :class:`BulkWriteOperation` instance, used to add
             update and remove operations to this bulk operation.
+
+        .. versionchanged:: 3.4
+           Added the `collation` option.
+
         """
         validate_is_mapping("selector", selector)
-        return BulkWriteOperation(selector, self.__bulk)
+        return BulkWriteOperation(selector, self.__bulk, collation)
 
     def insert(self, document):
         """Insert a single document.
@@ -626,5 +705,5 @@ class BulkOperationBuilder(object):
             execution.
         """
         if write_concern is not None:
-            validate_is_mapping("write_concern", write_concern)
-        return self.__bulk.execute(write_concern)
+            write_concern = WriteConcern(**write_concern)
+        return self.__bulk.execute(write_concern, session=None)
